<?php
namespace Teknisa\Libs\Service;

use Teknisa\Libs\Util\InstanceProvider;
use Teknisa\Libs\Util\Repositories\Migration as MigrationRepository;
use Teknisa\Libs\Util\Utilities;

class Migration {
    /** @var MigrationRepository $migrationRepository */
    private $migrationRepository;

    /**
     * @param MigrationRepository $migrationRepository
     */
    public function __construct(MigrationRepository $migrationRepository) {
        $this->migrationRepository = $migrationRepository;
    }

    /**
     * @param $productId
     * @param $nrOrg
     * @return array
     */
    public function getPendentMigrations($productId, $nrOrg) {
        try {
            $migrationInfo = array();
            $productPath   = Utilities::getProductBasePath();

            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            $repositories        = $this->migrationRepository->getMigrationInfo($productId, $customEntityManager);
            $masterRepository    = $this->migrationRepository->getMasterProductRepository($productId, $customEntityManager);
            $masterPath          = $masterRepository[0]['PATH'];

            foreach ($repositories as $repository) {
                $migDescription  = substr($repository['URL'], -50);
                $repositoryPath  = str_replace($masterPath, '', $repository['PATH']);
                $relativeMigPath = $repository['MIG_RELATIVE_PATH'];
                $scriptsPath     = $productPath . DIRECTORY_SEPARATOR . $repositoryPath . DIRECTORY_SEPARATOR . $relativeMigPath;

                $versionObj = $this->migrationRepository->getLastAppliedMigration($migDescription, $nrOrg, $customEntityManager);
                $lastFile   = null;

                if (!empty($versionObj)) {
                    $lastFile = $versionObj[0]['NRVEROBJBD'];
                }

                $initialMigrationNumber = $repository['INITIAL_MIGRATION_NUMBER'];
                $migFiles = $this->getMigrationFilesContents($lastFile, $scriptsPath, $initialMigrationNumber);

                $migrationInfo[] = array (
                    'REPOSITORY_ID'           => $repository['ID'],
                    'REPOSITORY_URL'          => $repository['URL'],
                    'REPOSITORY_NAME'         => $repository['NAME'],
                    'REPOSITORY_PATH'         => $repositoryPath,
                    'REPOSITORY_REL_PATH'     => $relativeMigPath,
                    'MIG_DESCRIPTION'         => $migDescription,
                    'MIG_LAST_FILE_NUMBER'    => $lastFile,
                    'MIG_HIGHER_FILE_NUMBER'  => $migFiles['HIGHER_FILE'],
                    'MIG_SMALLER_FILE_NUMBER' => $initialMigrationNumber ? $initialMigrationNumber : $migFiles['SMALL_FILE'],
                    'MIG_FILES_CONTENT'       => $lastFile ? $migFiles['FILES'] : array(),
                    'MIG_FULL_FILES_CONTENT'  => $lastFile ? $migFiles['FULL_FILES'] : array()
                );
            }

            return $migrationInfo;
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $lastFile
     * @param $scriptsPath
     * @param $initialMigrationNumber
     * @return array
     */
    private function getMigrationFilesContents($lastFile, $scriptsPath, $initialMigrationNumber) {
        $files = array_diff(scandir($scriptsPath), array('..', '.'));
        $higherFileNumber = '0';
        $smallFileNumber  = null;
        $filesWithContent = array();
        $fullFilesWithContent = array();

        foreach($files as $file) {
            if(strpos($file, 'down') === false) {
                if (strpos($file, '_')) {
                    $fileNumber = explode('_', $file)[0];
                } else {
                    $fileNumber = explode('.', $file)[0];
                }

                if (is_numeric($fileNumber)) {
                    if ($lastFile && $fileNumber > $lastFile) {
                        $this->addNotEmptyContents($scriptsPath, $file, $filesWithContent);
                        $this->addNotEmptyContents($scriptsPath, $file, $fullFilesWithContent);
                        $this->setHigherAndSmallFileNumber($fileNumber, $higherFileNumber, $smallFileNumber, true);
                    } else if($lastFile && $initialMigrationNumber && $fileNumber > $initialMigrationNumber) {
                        $this->addNotEmptyContents($scriptsPath, $file, $fullFilesWithContent);
                        $this->setHigherAndSmallFileNumber($fileNumber, $higherFileNumber, $smallFileNumber, true);
                    } else {
                        $this->setHigherAndSmallFileNumber($fileNumber, $higherFileNumber, $smallFileNumber, !$initialMigrationNumber);
                    }
                }
            }
        }

        return array(
            'FILES'       => $filesWithContent,
            'FULL_FILES'  => $fullFilesWithContent,
            'HIGHER_FILE' => $higherFileNumber,
            'SMALL_FILE'  => $smallFileNumber
        );
    }

    /**
     * @param $scriptsPath
     * @param $file
     * @param $filesWithContent
     */
    private function addNotEmptyContents($scriptsPath, $file, &$filesWithContent) {
        $fullContents = file_get_contents($scriptsPath . DIRECTORY_SEPARATOR . $file);
        $notEmptyContents = array();

        foreach (explode('/', $fullContents) as $contents) {
            if (!empty($contents)) {
                $notEmptyContents[] = trim($contents);
            }
        }

        $filesWithContent[] = array(
            'FILE_NAME' => $file,
            'FILE_CONTENTS' => $notEmptyContents
        );
    }

    /**
     * @param $fileNumber
     * @param $higherFileNumber
     * @param $smallFileNumber
     * @param $setSmall
     */
    private function setHigherAndSmallFileNumber($fileNumber, &$higherFileNumber, &$smallFileNumber, $setSmall) {
        if($fileNumber > $higherFileNumber) {
            $higherFileNumber = $fileNumber;
        }

        if((!$smallFileNumber || $fileNumber < $smallFileNumber) && $setSmall) {
            $smallFileNumber = $fileNumber;
        }
    }

    /**
     * @param $script
     * @param $nrOrg
     * @return array
     * @throws \Exception
     */
    public function applyMigration($script, $nrOrg) {
        try {
            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            $result = $this->migrationRepository->executeScript($script, $customEntityManager);
            return $this->errorTreatmentOnSQLExecution($result);
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $nmObject
     * @param $fileName
     * @param $oldFile
     * @param $nrOrg
     * @param $cdOperator
     * @return array
     */
    public function saveLastSuccessMigration($nmObject, $fileName, $oldFile, $nrOrg, $cdOperator) {
        try {
            $customEntityManager = InstanceProvider::getConnectionCustom()->updateVpdConnection($nrOrg);
            if($oldFile) {
                $result = $this->migrationRepository->updateLastMigration($nmObject, $fileName, $nrOrg, $cdOperator, $customEntityManager);
            } else {
                $result = $this->migrationRepository->insertLastMigration($nmObject, $fileName, $nrOrg, $cdOperator, $customEntityManager);
            }
            return $this->errorTreatmentOnSQLExecution($result);
        } catch (\Exception $e) {
            return $this->errorTreatmentOnSQLExecution($e);
        }
    }

    /**
     * @param $result
     * @return array
     */
    private function errorTreatmentOnSQLExecution($result) {
        $error = $rowCount = null;

        if($result instanceof \Exception) {
            /** @var \Exception $result */
            $error = $result->getMessage();
        } else {
            /** @var \Doctrine\DBAL\Driver\Statement $result */
            $errorInfo = $result->errorInfo();
            if(isset($errorInfo['message']) && !empty($errorInfo['message'])) {
                $error = $errorInfo['message'];
            }
            $rowCount = $result->rowCount();
        }
        
        if(strpos($error, 'ORA-') !== false) {
            $error = 'ORA-' . explode('ORA-', $error)[1];
        }
        
        return array(
            'ROW_COUNT' => $rowCount,
            'ERROR'     => $error
        );
    }
}